// Custom mutex and reader-writer lock implementations. Link with this module
// statically to override the versions provided by the system.
//
// Author: 15213-s16 Recitation Section C

#include "pthread.h"

#include <stdio.h>

//
// Mutex implementation (using a semaphore)
//

#ifdef USE_CUSTOM_MUTEX

int pthread_mutex_init(pthread_mutex_t *mutex,
		const pthread_mutexattr_t *mutexattr) {
	// Avoid generating an unused parameter warning.
	(void) mutexattr;

	// Print so we know we're using our own implementation instead of the
	// system one.
	puts("Initializing custom mutex");

	// A mutex is just the degenerate case of a semaphore.
	sem_init(&mutex->sem, 0, 1);
	return 0;
}

int pthread_mutex_destroy(pthread_mutex_t *mutex) {
	sem_destroy(&mutex->sem);
	return 0;
}

int pthread_mutex_lock(pthread_mutex_t *mutex) {
	sem_wait(&mutex->sem);
	return 0;
}

int pthread_mutex_unlock(pthread_mutex_t *mutex) {
	sem_post(&mutex->sem);
	return 0;
}

#endif

//
// Reader-writer lock implementation (using two mutexes)
//

#ifdef USE_CUSTOM_RWLOCK

int pthread_rwlock_init(pthread_rwlock_t *restrict rwlock,
		const pthread_rwlockattr_t *restrict attr) {
	// Avoid generating an unused parameter warning.
	(void) attr;

	// Print so we know we're using our own implementation instead of the
	// system one.
	puts("Initializing custom rwlock");

	pthread_mutex_init(&rwlock->lck, NULL);
	rwlock->ctr = 0;
	pthread_mutex_init(&rwlock->excl, NULL);
	return 0;
}

int pthread_rwlock_destroy(pthread_rwlock_t *rwlock) {
	pthread_mutex_destroy(&rwlock->lck);
	pthread_mutex_destroy(&rwlock->excl);
	return 0;
}

int pthread_rwlock_rdlock(pthread_rwlock_t *rwlock) {
	// We'll be modifying ctr, so we need to lock it.
	pthread_mutex_lock(&rwlock->lck);

	// Check whether we're the first reader. Because this is done while
	// holding lck, it's impossible for multiple readers to believe they're
	// the first one.
	if(rwlock->ctr == 0)
		// Just like the writers, the readers have exclusive access to
		// the resource; the difference is that this access is
		// collective in the case of the readers. As such, they delegate
		// one reader---the first to enter its critical section---to
		// take the exclusive lock on behalf of the entire group. To
		// prevent deadlock, it's important that this lock is later
		// released when the last reader unlocks the rwlock.
		pthread_mutex_lock(&rwlock->excl);

	// Record that one more reader is entering its critical section.
	++rwlock->ctr;
	pthread_mutex_unlock(&rwlock->lck);
	return 0;
}

int pthread_rwlock_wrlock(pthread_rwlock_t *rwlock) {
	// *Each* writer requires exclusive access to the resource. This call
	// will block until all readers *and* other writers have released the
	// reader-writer lock.
	pthread_mutex_lock(&rwlock->excl);
	return 0;
}

int pthread_rwlock_unlock(pthread_rwlock_t *rwlock) {
	if(rwlock->ctr == 0)
		// There are no readers, so we're a writer! Release our
		// exclusive access to the resource.
		pthread_mutex_unlock(&rwlock->excl);
	else {
		// We're a reader! Lock the ctr so we're ready to decrement it.
		pthread_mutex_lock(&rwlock->lck);

		// There's one fewer reader in a critical section.
		if(--rwlock->ctr == 0)
			// We were the last reader, so release the readers'
			// collective exclusive access to the resource.
			pthread_mutex_unlock(&rwlock->excl);

		pthread_mutex_unlock(&rwlock->lck);
	}
	return 0;
}

#endif
