// Specification for mutexes (using a semaphore) and a reader-writer locks
// (using two mutexes). Include this file in lieu of the pthread.h provided by
// the system, and optionally replace the system implementations with your own
// by linking in your own definitions of the functions declared herein.
//
// Author: Sol Boucher <sboucher@cmu.edu>

#ifndef PTHREAD_H_
#define PTHREAD_H_

// Change the names of the pthread types and function declarations in the system
// pthreads header. This will allow us to redefine them here without receiving
// duplicate/conflicting definition errors from the compiler.
#define pthread_mutex_t internal_pthread_mutex_t
#define pthread_mutex_init internal_pthread_mutex_init
#define pthread_mutex_destroy internal_pthread_mutex_destroy
#define pthread_mutex_lock internal_pthread_mutex_lock
#define pthread_mutex_unlock internal_pthread_mutex_unlock

#define pthread_rwlock_t internal_pthread_rwlock_t
#define pthread_rwlock_init internal_pthread_rwlock_init
#define pthread_rwlock_destroy internal_pthread_rwlock_destroy
#define pthread_rwlock_rdlock internal_pthread_rwlock_rdlock
#define pthread_rwlock_wrlock internal_pthread_rwlock_wrlock
#define pthread_rwlock_unlock internal_pthread_rwlock_unlock

// Include the system pthread header, which we'll need for pthread_create(),
// pthread_join(), and the other threading functions.
#include <pthread.h>

// End the name replacements so we'll be able to add our own declarations.
#undef pthread_rwlock_t
#undef pthread_rwlock_init
#undef pthread_rwlock_destroy
#undef pthread_rwlock_rdlock
#undef pthread_rwlock_wrlock
#undef pthread_rwlock_unlock

#undef pthread_mutex_t
#undef pthread_mutex_init
#undef pthread_mutex_destroy
#undef pthread_mutex_lock
#undef pthread_mutex_unlock

#include "semaphore.h"

//
// Mutex specification
//

typedef struct {
	// Implemented using a semaphore
	sem_t sem;
	// Expand the structure to be larger than the system one. This allows
	// linking against the stock pthreads implementation without having it
	// write beyond the end of pthread_mutex_t structures.
	char fill[sizeof(internal_pthread_mutex_t)];
} pthread_mutex_t;

int pthread_mutex_init(pthread_mutex_t *mutex,
		const pthread_mutexattr_t *mutexattr);
int pthread_mutex_destroy(pthread_mutex_t *mutex);
int pthread_mutex_lock(pthread_mutex_t *mutex);
int pthread_mutex_unlock(pthread_mutex_t *mutex);

//
// Reader-writer lock specification
//

typedef struct {
	// Guard ctr modifications (increments and decrements are non-atomic).
	pthread_mutex_t lck;
	// The number of read locks currently held
	unsigned ctr;
	// Exclusive access (to the resource we're protecting w/ the rwlock)
	pthread_mutex_t excl;
	// Expand to larger than the system structure to maintain compatibility.
	char fill[sizeof(internal_pthread_rwlock_t)];
} pthread_rwlock_t;

int pthread_rwlock_init(pthread_rwlock_t *restrict rwlock,
		const pthread_rwlockattr_t *restrict attr);
int pthread_rwlock_destroy(pthread_rwlock_t *rwlock);
int pthread_rwlock_rdlock(pthread_rwlock_t *rwlock);
int pthread_rwlock_wrlock(pthread_rwlock_t *rwlock);
int pthread_rwlock_unlock(pthread_rwlock_t *rwlock);

#endif
